@@ -11,11 +11,17 @@ from . import tenbin
 from .dataset import (
     Dataset,
     ChoppedDataset,
+    ResizedDataset,
+    Pipeline,
+    SampleIterator,
     reraise_exception,
     ignore_and_continue,
     warn_and_continue,
     ignore_and_stop,
     warn_and_stop,
+    worker_urls,
+    all_urls
 )
 from .autodecode import default_handlers
 from .writer import ShardWriter, TarWriter
+from .multi import MultiDataset
