@@ -26,10 +26,9 @@ from builtins import range
 import braceexpand
 from torch.utils.data import IterableDataset
 
-from . import gopen
-from . import filters
-from .checks import checkcallable
+from . import filters, gopen
 
+# from functools import wraps
 trace = False
 
 debug_dataset = os.environ.get("WDS_DEBUG", 0)
@@ -71,6 +70,45 @@ def warn_and_stop(exn):
     return False
 
 
+def add_hook(fs, f):
+    assert callable(f)
+    if fs is None:
+        return [f]
+    if isinstance(fs, list):
+        return fs + [f]
+    assert callable(fs)
+    return [fs, f]
+
+
+def call_hook(fs, *args, **kw):
+    if fs is None:
+        return
+    if not isinstance(fs, list):
+        fs = [fs]
+    for f in fs:
+        f(*args, **kw)
+
+
+def identity(x):
+    return x
+
+
+def do_nothing(*args, **kw):
+    """Do nothing function."""
+    pass
+
+
+class Shuffler:
+    """Make a shuffle function (avoid nesting for pickle)."""
+    def __init__(self, rng):
+        self.rng = rng
+
+    def __call__(self, lst):
+        lst = list(lst)
+        self.rng.shuffle(lst)
+        return lst
+
+
 def maybe_collect():
     """Running in notebooks, we tend to run out of memory due to
     weak references, and the collector doesn't seem to get triggered
@@ -112,7 +150,7 @@ def valid_sample(sample):
     )
 
 
-def group_by_keys(keys=base_plus_ext, lcase=True, suffixes=None):
+def group_by_keys_(data, keys=base_plus_ext, lcase=True, suffixes=None):
     """Returns function over iterator that groups key, value pairs into samples.
 
     keys: function that splits the key into key and extension (base_plus_ext)
@@ -120,34 +158,34 @@ def group_by_keys(keys=base_plus_ext, lcase=True, suffixes=None):
 
     """
 
-    def iterator(data):
-        current_sample = None
-        for fname, value in data:
-            prefix, suffix = keys(fname)
-            if trace:
-                print(
-                    prefix,
-                    suffix,
-                    current_sample.keys() if isinstance(current_sample, dict) else None,
-                )
-            if prefix is None:
-                continue
-            if lcase:
-                suffix = suffix.lower()
-            if current_sample is None or prefix != current_sample["__key__"]:
-                if valid_sample(current_sample):
-                    yield current_sample
-                current_sample = dict(__key__=prefix)
-            if suffix in current_sample:
-                raise ValueError(
-                    f"{fname}: duplicate file name in tar file {suffix} {current_sample.keys()}"
-                )
-            if suffixes is None or suffix in suffixes:
-                current_sample[suffix] = value
-        if valid_sample(current_sample):
-            yield current_sample
+    current_sample = None
+    for fname, value in data:
+        prefix, suffix = keys(fname)
+        if trace:
+            print(
+                prefix,
+                suffix,
+                current_sample.keys() if isinstance(current_sample, dict) else None,
+            )
+        if prefix is None:
+            continue
+        if lcase:
+            suffix = suffix.lower()
+        if current_sample is None or prefix != current_sample["__key__"]:
+            if valid_sample(current_sample):
+                yield current_sample
+            current_sample = dict(__key__=prefix)
+        if suffix in current_sample:
+            raise ValueError(
+                f"{fname}: duplicate file name in tar file {suffix} {current_sample.keys()}"
+            )
+        if suffixes is None or suffix in suffixes:
+            current_sample[suffix] = value
+    if valid_sample(current_sample):
+        yield current_sample
 
-    return iterator
+
+group_by_keys = filters.Curried(group_by_keys_)
 
 
 def tardata(fileobj, skip_meta=r"__[^/]*__($|/)", handler=reraise_exception):
@@ -187,112 +225,49 @@ def tardata(fileobj, skip_meta=r"__[^/]*__($|/)", handler=reraise_exception):
         handler(exn)
 
 
-class Dataset(IterableDataset):
-    """Iterate over sharded datasets.
-
-    :param urls: shard spec or list of shards
-    :param prepare_for_worker: callable called in each worker before anything else is done
+class Pipeline:
+    """Simple fluid pipeline builder.
 
+    This is a convenience class that allows common processing
+    pipelines to be built up with a fluid interface.
     """
 
-    def __init__(
-        self,
-        urls,
-        *,
-        keys=base_plus_ext,
-        suffixes=None,
-        length=None,
-        epochs=1,
-        opener=gopen.reader,
-        handler=reraise_exception,
-        shuffle=False,
-        prepare_for_worker=True,
-        initial_pipeline=None,
-    ):
-        if isinstance(opener, str):
-            self.opener = gopen.command_pipe(opener)
-        elif callable(opener):
-            self.opener = opener
-        else:
-            raise ValueError(f"{opener}: must be either str or callable")
-        checkcallable(self.opener)
-        if isinstance(urls, str):
-            urls = list(braceexpand.braceexpand(urls))
-        self.full_urls = self.urls = urls
-        self.length = length
-        self.epochs = epochs
-        self.keys = keys
-        self.suffixes = suffixes
-        self.subset = None
-        self.do_shuffle = shuffle
-        self.handler = handler
-        self.rng = random
-        if prepare_for_worker is True:
-            self.prepare_for_worker = self.shard_selection
-        elif prepare_for_worker is False:
-            self.prepare_for_worker = lambda: None
-        else:
-            self.prepare_for_worker = prepare_for_worker
-        self.pipeline = (
-            initial_pipeline if initial_pipeline is not None else [group_by_keys()]
-        )
-
-    def __len__(self):
-        return self.length
-
-    def shard_selection(self):
-        """Contains the logic for self.subset shard selection."""
-        import torch
-
-        index, total = None, None
-        if self.subset is not None:
-            index, total = self.subset  # skipcq: PYL-E0633
-        else:
-            worker_info = torch.utils.data.get_worker_info()
-            if worker_info is not None:
-                index = worker_info.id
-                total = worker_info.num_workers
-        if total is None:
-            self.urls = self.full_urls
-            return
-        if index == 0 and len(self.full_urls) < total:
-            warnings.warn(f"num_workers {total} > num_shards {len(self.full_urls)}")
-        self.urls = self.full_urls[index::total]
-
-    def raw_iter(self):
-        """Iterate over samples."""
-        self.prepare_for_worker()
-        if self.do_shuffle:
-            self.rng.shuffle(self.urls)
-        self.sample = 0
-        urls = self.urls
-        count = 0
-        for epoch in range(self.epochs):
-            for url in urls:
-                stream = None
-                with self.opener(url) as stream:
-                    files_of_archive = tardata(stream, handler=self.handler)
-                    for fname, content in files_of_archive:
-                        yield fname, content
-                        count += 1
-                    maybe_collect()
-
-    def __iter__(self):
-        return filters.pipeline(self.raw_iter(), *self.pipeline)
+    def __init__(self):
+        self.pipeline = []
+        self.rng = None
+        self.shard_shuffle = None
 
     def pipe(self, stage):
         """Add a pipline stage (a function taking an iterator and returning another iterator)."""
         self.pipeline.append(stage)
         return self
 
-    def shuffle(self, size, **kw):
+    def batched(
+        self, batchsize, partial=True,
+    ):
+        self.pipeline.append(filters.batched(batchsize=batchsize, partial=True,))
+        return self
+
+    def unbatched(self):
+        self.pipeline.append(filters.unbatched())
+        return self
+
+    def shuffle(self, size, rng=None, **kw):
         """Shuffle the data."""
         if size == 0:
             return self
-        self.do_shuffle = True
-        self.pipeline.append(filters.shuffle(size, rng=self.rng, **kw))
+        if rng is None:
+            rng = random.Random()
+        self.rng = rng
+        self.reseed_hook = self.reseed_rng
+        self.shard_shuffle = Shuffler(rng)
+        self.pipeline.append(filters.shuffle(size, rng=rng, **kw))
         return self
 
+    def reseed_rng(self):
+        seed = random.SystemRandom().random()
+        self.rng.seed(seed)
+
     def decode(self, decoder="rgb", handler=reraise_exception):
         """Decode the data with the given decoder."""
         self.pipeline.append(filters.decode(decoder, handler=handler))
@@ -328,10 +303,143 @@ class Dataset(IterableDataset):
         self.pipeline.append(filters.map_tuple(*args, handler=handler))
         return self
 
-    # TODO: add batching here
+
+def make_opener(open_fn):
+    if isinstance(open_fn, str):
+        return gopen.command_pipe(open_fn)
+    elif callable(open_fn):
+        return open_fn
+    else:
+        raise ValueError(f"{open_fn}: must be either str or callable")
 
 
-class ChoppedDataset(IterableDataset):
+class SampleIterator(Pipeline):
+    """Iterates over the samples of webdatasets using a given processing pipeline."""
+
+    def __init__(
+        self, initial_pipeline=None, tarhandler=reraise_exception, open_fn=gopen.reader
+    ):
+        Pipeline.__init__(self)
+        self.open_fn = make_opener(open_fn)
+        self.tarhandler = tarhandler
+        self.shard_hook = do_nothing
+        if initial_pipeline is None:
+            initial_pipeline = [group_by_keys()]
+        for stage in initial_pipeline:
+            self.pipe(stage)
+
+    def raw_samples(self, urls):
+        assert isinstance(urls, list)
+        for url in urls:
+            self.shard_hook()
+            stream = None
+            try:
+                with self.open_fn(url) as stream:
+                    files_of_archive = tardata(stream, handler=self.tarhandler)
+                    for fname, content in files_of_archive:
+                        yield fname, content
+                    maybe_collect()
+            except Exception as exn:
+                if self.tarhandler(exn):
+                    continue
+                else:
+                    break
+
+    def samples(self, urls):
+        if isinstance(urls, str):
+            urls = [urls]
+        assert isinstance(urls, list)
+        self.sample_urls = urls
+        source = self.raw_samples(urls)
+        return filters.pipeline(source, *self.pipeline)
+
+
+def all_urls(urls):
+    """Returns all URLs.
+
+    Used as a shard selection function in Dataset."""
+    assert isinstance(urls, list)
+    assert isinstance(urls[0], str)
+    return urls
+
+
+def worker_urls(urls):
+    """Selects a subset of urls based on Torch get_worker_info.
+
+    Used as a shard selection function in Dataset."""
+    import torch
+
+    assert isinstance(urls, list)
+    assert isinstance(urls[0], str)
+
+    worker_info = torch.utils.data.get_worker_info()
+    if worker_info is not None:
+        wid = worker_info.id
+        num_workers = worker_info.num_workers
+        if wid == 0 and len(urls) < num_workers:
+            warnings.warn(f"num_workers {num_workers} > num_shards {len(urls)}")
+        return urls[wid::num_workers]
+    else:
+        return urls
+
+
+class Dataset(IterableDataset, SampleIterator):
+    """Iterate over sharded datasets.
+
+    This class combines several function: it is a container for a list of
+    shards, it is a container for a processing pipelines, and it handles
+    some bookkeeping related to DataLoader.
+    """
+
+    def __init__(
+        self,
+        urls,
+        *,
+        length=None,
+        open_fn=gopen.reader,
+        handler=reraise_exception,
+        tarhandler=None,
+        prepare_for_worker=True,
+        initial_pipeline=None,
+        shard_selection=worker_urls,
+    ):
+        tarhandler = handler if tarhandler is None else tarhandler
+        IterableDataset.__init__(self)
+        SampleIterator.__init__(
+            self,
+            initial_pipeline=initial_pipeline,
+            tarhandler=tarhandler,
+            open_fn=open_fn,
+        )
+        if isinstance(urls, str):
+            urls = list(braceexpand.braceexpand(urls))
+        self.urls = urls
+        self.length = length
+        self.handler = handler
+        self.total = 0
+        self.reseed_hook = do_nothing
+        self.node_selection = identity
+        self.shard_selection = shard_selection
+        self.shard_shuffle = identity
+
+    def __len__(self):
+        """Return the nominal length of the dataset."""
+        return self.length
+
+    def shard_fn(self):
+        urls = list(self.urls)
+        self.reseed_hook()
+        urls = self.node_selection(urls)
+        urls = self.shard_selection(urls)
+        urls = self.shard_shuffle(urls)
+        return urls
+
+    def __iter__(self):
+        urls = self.shard_fn()
+        return self.samples(urls)
+
+
+class ResizedDataset(IterableDataset):
     """Change the actual and nominal length of an IterableDataset.
 
     :param dataset: IterableDataset
@@ -339,24 +447,32 @@ class ChoppedDataset(IterableDataset):
     :param nominal: nominal length of dataset (if different from declared)
 
     This will continuously iterate through the original dataset, but
-    impose new eopch boundaries at the given length/nominal.
+    impose new epoch boundaries at the given length/nominal.
     This exists mainly as a workaround for the odd logic in DataLoader.
     It is also useful for choosing smaller nominal epoch sizes with
     very large datasets.
 
     """
+
     def __init__(self, dataset, length=None, nominal=None):
         self.dataset = dataset
         if length is None:
             length = len(dataset)
         self.length = length
         self.nominal = self.length if nominal is None else nominal
-        self.source = iter(dataset)
+        self.source = None
 
     def __len__(self):
         return self.nominal
 
+    def __getstate__(self):
+        result = dict(self.__dict__)
+        result["source"] = None
+        return result
+
     def __iter__(self):
+        if self.source is None:
+            self.source = iter(self.dataset)
         for i in range(self.length):
             try:
                 sample = next(self.source)
@@ -364,3 +480,6 @@ class ChoppedDataset(IterableDataset):
                 self.source = iter(self.dataset)
                 sample = next(self.source)
             yield sample
+
+
+ChoppedDataset = ResizedDataset
