@@ -12,29 +12,61 @@ or over HTTP connections.
 
 __all__ = "WebDataset tariterator default_handlers imagehandler".split()
 
-import sys
 import random
-from functools import reduce, wraps
+import sys
+from functools import reduce
+
+import numpy as np
 
-from .checks import checktype
 from . import autodecode
+from .checks import checktype
+
+try:
+    from torch import Tensor as TorchTensor
+except ModuleNotFoundError:
+
+    class TorchTensor:
+        pass
+
+
+class Curried2(object):
+    """Helper class for currying pipeline stages.
+
+    We use this roundabout construct becauce it can be pickled."""
+
+    def __init__(self, f, *args, **kw):
+        self.f = f
+        self.args = args
+        self.kw = kw
+
+    def __call__(self, data):
+        return self.f(data, *self.args, **self.kw)
+
+    def __str__(self):
+        return f"<{self.f.__name__} {self.args} {self.kw}>"
+
+    def __repr__(self):
+        return f"<{self.f.__name__} {self.args} {self.kw}>"
+
+
+class Curried(object):
+    """Helper class for currying pipeline stages.
+
+    We use this roundabout construct becauce it can be pickled."""
+
+    def __init__(self, f):
+        self.f = f
+
+    def __call__(self, *args, **kw):
+        return Curried2(self.f, *args, **kw)
 
 
 def reraise_exception(exn):
     raise exn
 
 
-def curried(f):
-    """A decorator for currying functions in the first argument."""
-
-    @wraps(f)
-    def wrapper(*args, **kw):
-        def g(x):
-            return f(x, *args, **kw)
-
-        return g
-
-    return wrapper
+def identity(x):
+    return x
 
 
 def compose2(f, g):
@@ -128,8 +160,7 @@ def transformer(transformers):
 #             yield [getfirst(sample, f) for f in fields]
 
 
-@curried
-def map_stream(data, f=None, handler=reraise_exception):
+def map_stream_(data, f=None, handler=reraise_exception):
     """Map entire samples using the given function.
 
     data: iterator
@@ -156,8 +187,10 @@ def map_stream(data, f=None, handler=reraise_exception):
         yield result
 
 
-@curried
-def info(data, fmt=None, n=3, every=-1, width=50, stream=sys.stderr, name=""):
+map_stream = Curried(map_stream_)
+
+
+def info_(data, fmt=None, n=3, every=-1, width=50, stream=sys.stderr, name=""):
     for i, sample in enumerate(data):
         if i < n or (every > 0 and (i + 1) % every == 0):
             if fmt is None:
@@ -169,8 +202,10 @@ def info(data, fmt=None, n=3, every=-1, width=50, stream=sys.stderr, name=""):
         yield sample
 
 
-@curried
-def shuffle(data, bufsize=1000, initial=100, rng=random):
+info = Curried(info_)
+
+
+def shuffle_(data, bufsize=1000, initial=100, rng=random):
     """Shuffle the data in the stream.
 
     This uses a buffer of size `bufsize`. Shuffling at
@@ -203,133 +238,209 @@ def shuffle(data, bufsize=1000, initial=100, rng=random):
         yield sample
 
 
-@curried
-def select(data, predicate):
+shuffle = Curried(shuffle_)
+
+
+def select_(data, predicate):
     for sample in data:
         if predicate(sample):
             yield sample
 
 
-def decode(decoder="rgb", handler=reraise_exception):
+select = Curried(select_)
+
+
+def decode_(data, decoder="rgb", handler=reraise_exception):
     f = autodecode.make_decoder(decoder)
 
-    def stage(data):
-        for sample in data:
-            assert isinstance(sample, dict), sample
-            try:
-                decoded = f(sample)
-            except Exception as exn:  # skipcq: PYL-W0703
-                if handler(exn):
-                    continue
-                else:
-                    break
-            yield decoded
-
-    return stage
-
-
-def map(f, handler=reraise_exception):
-    def stage(data):
-        for sample in data:
-            try:
-                result = f(sample)
-            except Exception as exn:
-                if handler(exn):
-                    continue
-                else:
-                    break
-            if isinstance(sample, dict) and isinstance(result, dict):
-                result["__key__"] = sample.get("__key__")
-            yield result
-
-    return stage
-
-
-def rename(handler=reraise_exception, **kw):
-    def stage(data):
-        for sample in data:
-            try:
-                yield {
-                    k: getfirst(sample, v, missing_is_error=True) for k, v in kw.items()
-                }
-            except Exception as exn:
-                if handler(exn):
-                    continue
-                else:
-                    break
-
-    return stage
-
-
-def associate(associator, **kw):
-    def stage(data):
-        for sample in data:
-            if callable(associator):
-                extra = associator(sample["__key__"])
+    for sample in data:
+        assert isinstance(sample, dict), sample
+        try:
+            decoded = f(sample)
+        except Exception as exn:  # skipcq: PYL-W0703
+            if handler(exn):
+                continue
             else:
-                extra = associator.get(sample["__key__"], {})
-            sample.update(extra)  # destructive
-            yield sample
-
-    return stage
+                break
+        yield decoded
 
 
-def map_dict(handler=reraise_exception, **kw):
+decode = Curried(decode_)
+
+
+def map_(data, f, handler=reraise_exception):
+    for sample in data:
+        try:
+            result = f(sample)
+        except Exception as exn:
+            if handler(exn):
+                continue
+            else:
+                break
+        if isinstance(sample, dict) and isinstance(result, dict):
+            result["__key__"] = sample.get("__key__")
+        yield result
+
+
+map = Curried(map_)
+
+
+def rename_(data, handler=reraise_exception, **kw):
+    for sample in data:
+        try:
+            yield {k: getfirst(sample, v, missing_is_error=True) for k, v in kw.items()}
+        except Exception as exn:
+            if handler(exn):
+                continue
+            else:
+                break
+
+
+rename = Curried(rename_)
+
+
+def associate_(data, associator, **kw):
+    for sample in data:
+        if callable(associator):
+            extra = associator(sample["__key__"])
+        else:
+            extra = associator.get(sample["__key__"], {})
+        sample.update(extra)  # destructive
+        yield sample
+
+
+associate = Curried(associate_)
+
+
+def map_dict_(data, handler=reraise_exception, **kw):
     assert len(list(kw.keys())) > 0
-    for f in kw.values():
-        assert callable(f)
+    for key, f in kw.items():
+        assert callable(f), (key, f)
 
-    def stage(data):
-        for sample in data:
-            assert isinstance(sample, dict)
-            try:
-                for k, f in kw.items():
-                    sample[k] = f(sample[k])
-            except Exception as exn:
-                if handler(exn):
-                    continue
-                else:
-                    break
-            yield sample
-
-    return stage
+    for sample in data:
+        assert isinstance(sample, dict)
+        try:
+            for k, f in kw.items():
+                sample[k] = f(sample[k])
+        except Exception as exn:
+            if handler(exn):
+                continue
+            else:
+                break
+        yield sample
 
 
-def to_tuple(*args, handler=reraise_exception):
+map_dict = Curried(map_dict_)
+
+
+def to_tuple_(data, *args, handler=reraise_exception):
     if len(args) == 1 and isinstance(args[0], str) and " " in args[0]:
         args = args[0].split()
 
-    def stage(data):
-        for sample in data:
-            try:
-                yield tuple([getfirst(sample, f, missing_is_error=True) for f in args])
-            except Exception as exn:
-                if handler(exn):
-                    continue
-                else:
-                    break
-
-    return stage
+    for sample in data:
+        try:
+            yield tuple([getfirst(sample, f, missing_is_error=True) for f in args])
+        except Exception as exn:
+            if handler(exn):
+                continue
+            else:
+                break
 
 
-def map_tuple(*args, handler=reraise_exception):
-    def stage(data):
-        for f in args:
-            assert callable(f)
-        for sample in data:
-            assert isinstance(sample, (list, tuple))
-            assert len(args) == len(
-                sample
-            ), f"len(args) {len(args)} != len(sample) {len(sample)}"
-            sample = list(sample)
-            try:
-                for i in range(min(len(args), len(sample))):
-                    sample[i] = args[i](sample[i])
-            except Exception as exn:
-                if handler(exn):
-                    continue
-                else:
-                    break
-            yield tuple(sample)
+to_tuple = Curried(to_tuple_)
 
-    return stage
+
+def map_tuple_(data, *args, handler=reraise_exception):
+    for f in args:
+        assert callable(f), f
+    for sample in data:
+        assert isinstance(sample, (list, tuple))
+        assert len(args) == len(
+            sample
+        ), f"len(args) {len(args)} != len(sample) {len(sample)}"
+        sample = list(sample)
+        try:
+            for i in range(min(len(args), len(sample))):
+                sample[i] = args[i](sample[i])
+        except Exception as exn:
+            if handler(exn):
+                continue
+            else:
+                break
+        yield tuple(sample)
+
+
+map_tuple = Curried(map_tuple_)
+
+
+def default_collation_fn(samples, combine_tensors=True, combine_scalars=True):
+    """Take a collection of samples (dictionaries) and create a batch.
+
+    If `tensors` is True, `ndarray` objects are combined into
+    tensor batches.
+
+    :param dict samples: list of samples
+    :param bool tensors: whether to turn lists of ndarrays into a single ndarray
+    :returns: single sample consisting of a batch
+    :rtype: dict
+
+    """
+    batched = list(zip(*samples))
+    result = []
+    for b in batched:
+        if isinstance(b[0], (int, float)):
+            if combine_scalars:
+                b = np.array(list(b))
+        elif isinstance(b[0], TorchTensor):
+            if combine_tensors:
+                import torch
+                b = torch.stack(list(b))
+        elif isinstance(b[0], np.ndarray):
+            if combine_tensors:
+                b = np.array(list(b))
+        else:
+            b = list(b)
+        result.append(b)
+    return result
+
+
+def batched_(
+    data,
+    batchsize=20,
+    collation_fn=default_collation_fn,
+    partial=True,
+):
+    """Create batches of the given size.
+
+    :param data: iterator
+    :param batchsize: target batch size
+    :param tensors: automatically batch lists of ndarrays into ndarrays
+    :param partial: return partial batches
+    :returns: iterator
+
+    """
+
+    batch = []
+    for sample in data:
+        if len(batch) >= batchsize:
+            yield collation_fn(batch)
+            batch = []
+        batch.append(sample)
+    if len(batch) == 0:
+        return
+    elif len(batch) == batchsize or partial:
+        yield collation_fn(batch)
+
+
+batched = Curried(batched_)
+
+
+def unbatched_(data):
+    for sample in data:
+        assert isinstance(sample, (tuple, list)), sample
+        assert len(sample) > 0
+        for i in range(len(sample[0])):
+            yield tuple(x[i] for x in sample)
+
+
+unbatched = Curried(unbatched_)
